from django.db import models
from django.conf import settings
from django.contrib.auth.models import User
from django.utils import timezone
from datetime import timedelta
from image.models import Disk
from interface_openstack.implementation.vm.instance import (
        OSVirtualMachineManager
      )
import logging

logger = logging.getLogger(__name__)

ACCESS_METHODS = tuple(
    [(key, val[0]) for key, val in settings.VM_ACCESS_PROTOCOLS.items()]
)

interface = OSVirtualMachineManager(settings.CONNECTION)

ACTIONS = {
    "start": interface.start_vm,
    "stop": interface.stop_vm,
    "suspend": interface.suspend_vm,
    "wake_up": interface.wake_up_vm,
    "reset": interface.reset_vm,
    "reboot": interface.reboot_vm,
}


class Lease(models.Model):
    """ Users can use the virtual machine until the lease dates.
        After suspend interval the vm suspends and after delete it will be
        destroyed
    """
    name = models.CharField(blank=True, max_length=100)
    description = models.CharField(blank=True, max_length=100)
    suspend_interval_in_sec = models.IntegerField(blank=True, null=True)
    delete_interval_in_sec = models.IntegerField(blank=True, null=True)


class Flavor(models.Model):
    """ Flavors are reasource packages, contains all information about
        resources accociated with the virtual machine
    """
    name = models.CharField(blank=True, max_length=100)
    description = models.CharField(blank=True, max_length=200)
    remote_id = models.CharField(
            max_length=100, help_text="ID of the instance on the backend"
        )
    ram = models.IntegerField(blank=True, null=True)
    vcpu = models.IntegerField(blank=True, null=True)
    initial_disk = models.IntegerField(blank=True, null=True)
    priority = models.IntegerField(blank=True, null=True)

    @classmethod
    def create(cls, name, description, ram=0, vcpu=0,
               initial_disk=0, priority=0):
        try:
            remote_flavor = interface.create_flavor(name, ram, vcpu, initial_disk)

            flavor = cls(name=name, description=description,
                         remote_id=remote_flavor.id, ram=ram, vcpu=vcpu,
                         initial_disk=initial_disk, priority=priority)
            flavor.save()
            return flavor
        except Exception as e:
            logger.error(str(e))
            raise ValueError("Can't create Flavor in remote cloud.")

    def delete(self):
        interface = OSVirtualMachineManager(settings.CONNECTION)
        try:
            interface.delete_flavor(self.remote_id)
            super(Flavor, self).delete()
        except Exception as e:
            if e.OpenStackError:
                logger.error("Can not delete the flavor in remote cloud")


class Instance(models.Model):
    """Virtual machine instance.
    """
    from template.models import ImageTemplate

    name = models.CharField(max_length=100,
                            help_text="Human readable name of instance")

    remote_id = models.CharField(
        max_length=100, help_text="ID of the instance on the backend"
    )
    description = models.TextField(
        blank=True, help_text="The description of the instance"
    )
    access_method = models.CharField(
        max_length=10, choices=ACCESS_METHODS,
        help_text="Primary remote access method"
    )
    system = models.CharField(max_length=50, help_text="Operating system type")
    password = models.CharField(
        max_length=50, help_text="Original password of the instance"
    )
    time_of_suspend = models.DateTimeField(
        blank=True,
        help_text="After this point in time, the instance will be suspended"
    )
    time_of_delete = models.DateTimeField(
        blank=True,
        help_text="After this point in time, the instance will be deleted!"
    )
    deleted = models.BooleanField(
        help_text="Indicates if the instance is ready for garbage collection",
        default=False,
    )

    template = models.ForeignKey(ImageTemplate, related_name="vm", null=True,
                                 help_text="The base image of the vm",
                                 on_delete="DO_NOTHING")

    disks = models.ManyToManyField(Disk, related_name="instance",
                                   help_text="Disks attached to instance",
                                   verbose_name="disks")

    owner = models.ForeignKey(User, on_delete="CASCADE", null=True)
    flavor = models.ForeignKey(Flavor, help_text="Reasources given to the vm",
                               verbose_name="flavor", on_delete="CASCADE",
                               related_name='instances')
    lease = models.ForeignKey(Lease, on_delete="CASCADE",
                              related_name='instances')

    @classmethod
    def create(cls, lease, owner, flavor, template, remote_id, params):
        params["password"] = cls.generate_password()
        inst = cls(lease=lease, flavor=flavor, owner=owner,
                   remote_id=remote_id, template=template, **params)
        inst.full_clean()
        inst.save()

        logger.info("New instance created")
        return inst

    @classmethod
    def create_instance_from_template(cls, params, template, owner, lease,
                                      disks, networks, flavor):
        # TODO: attach disks when the remote instance created
        try:
            remote_id = interface.create_vm_from_template(params["name"],
                                                          template.image.remote_id,
                                                          flavor.remote_id,
                                                          networks,
                                                          )
            new_inst = cls.create(lease, owner, flavor, template,
                                  remote_id, params)
            return new_inst
        except Exception as e:
            logger.error(str(e))
            raise ValueError("Can't create Instance in remote cloud."
                             "Search the logs for more detail.")

    def clean(self, *args, **kwargs):
        self.time_of_suspend, self.time_of_delete = self.get_renew_times()
        super(Instance, self).clean(*args, **kwargs)

    def get_renew_times(self, lease=None):
        """Returns new suspend and delete times if renew would be called.
        """
        if lease is None:
            lease = self.lease
        return (
            timezone.now() + timedelta(
                                seconds=lease.suspend_interval_in_sec),
            timezone.now() + timedelta(
                                seconds=lease.delete_interval_in_sec)
            )

    def delete(self):
        try:
            interface.destroy_vm(self.remote_id)
            super(Instance, self).delete()
        except Exception as e:
            if e.OpenStackError:
                logger.error("Can not delete the instance in remote cloud")

    def execute_common_action(self, action):
        if ACTIONS[action]:
            return ACTIONS[action](self.remote_id)
        else:
            raise ValueError("This action is not supported!")

    def get_remote_instance(self):
        return interface.get_vm(self.remote_id)

    @classmethod
    def generate_password(self):
        return User.objects.make_random_password(
                allowed_chars='abcdefghijklmnopqrstuvwx'
                              'ABCDEFGHIJKLMNOPQRSTUVWX123456789')
